#!/usr/bin/env python

__version__ = '$Revision: 4791 $'.split()[1]
__date__ = '$Date: 2007-01-04 $'.split()[1]
__author__ = 'Kurt Schwehr'

__doc__='''

Normalize AIS messages so that each message takes exactly one line.

Hopefully should be able to handle streams with multiple receivers if thwy

Only deals with sqlite rightw now.

@requires: U{epydoc<http://epydoc.sourceforge.net/>} > 3.0alpha3
@requires: U{BitVector<http://cheeseshop.python.org/pypi/BitVector>}

@author: '''+__author__+'''
@version: ''' + __version__ +'''
@var __date__: Date of last svn commit
@undocumented: __version__ __author__ __doc__ parser
@status: under development
@license: Generated code has no license

@todo: report on the number of message fragments that got dropped
@todo: allow for a single receiver and no uscg station

@note: errors mostly seem to be comming from station r00370003 in Puerto Rico
'''

import sys
from decimal import Decimal
from BitVector import BitVector
import StringIO

from ais.nmea import isChecksumValid,checksumStr # Needed for checksums

def assembleAisNmeaMessages(infile=sys.stdin,outfile=sys.stdout,uscg=True,validateChecksum=True, verbose=True):
    '''
    Put together messages
    @param infile: file stream like object to read from
    @param outfile: some object that can take write messages for output

    '''
    o = outfile

    if not uscg:
	print 'Need to make a faster version that does not worry about the extra args and stations dict'
	assert False


    buffers = {} # Put partial messages in a queue by station so that they can be reassembled
    lineNum = 0
    for line in infile:
	line = line.strip()+'\n'  # Get rid of DOS issues - FIX: do I really want this line???
	lineNum += 1
	if line[3:6] not in ('VDM|VDO'):
	    o.write (line)  # Pass non AIS wireless message straight through
	    continue
	if validateChecksum:
	    if not isChecksumValid(line):
		print 'ERROR: Invalid checksum on line ',lineNum
		print '  ',line
	fields = line.split(',')
	
        if len(fields) < 6:
            sys.stderr.write('ERROR line '+str(lineNum)+': not enough fields in line...\n')
            sys.stderr.write('  '+line)
            continue


	totNumSentences = int(fields[1]) # Total nmea lines that compose this message 1..9 (numPackets)
	if 1 ==totNumSentences:
	    # Easy case
	    o.write (line)
	    continue

	sentenceNum = int(fields[2]) 	# Message sequence number 1..9 (packetNum)
	#seqId       = fields[3]		# Sequential message identifier 0..9
	#aisChanNum  = fields[4] 	# AIS channel number  A or B?
	payload     = fields[5] 	# AIS binary data encoded in whacky ways
	timestamp   = fields[-1].strip() 	# Seconds since Epoch UTC.  Always the last field

	station = None  # USCG Receive Stations
	for i in range(len(fields)-1,5,-1):
	    if 0<len(fields[i]) and 'r' == fields[i][0]:
		station = fields[i]
		break # Found it so ditch the for loop

        #if fields[3]==None: fields[3]=''
        #if fields[4]==None: fields[4]=''
        if None == station:
            sys.stderr.write('ERROR line '+str(lineNum)+': not enough fields in line... skipping\n')
            sys.stderr.write('  '+line)
            continue
            #sys.exit('I QUIT!')

        bufferSlot = station + fields[3] + fields[4]  # seqId and Channel make a unique stream
            

	newPacket = (payload,station,timestamp)

	if sentenceNum == 1:
	    buffers[bufferSlot] = [newPacket] # Overwrite any partials
	    continue

	if totNumSentences == sentenceNum:
	    # Finished a message
	    #print 'buffers',buffers
	    if bufferSlot not in buffers:
		if verbose: print 'Do not have the preceeding packets for line'
		if verbose: print '  ',line
		continue
	    buffers[bufferSlot].append(newPacket)
	    parts = buffers[bufferSlot] # Now have all the pieces
	    #if verbose: print type(buffers[bufferSlot]), buffers[bufferSlot]
	    del buffers[bufferSlot] # Clear out the used packets to save memory

	    # Sanity check
	    ok = True
	    for part in parts:
		if part[2] != timestamp:
		    print 'ERROR: timestamps not all the same for ',timestamp
		    print '  ** ',line
		    print '  parts:',parts
		    ok=False
		    break
	    if not ok: continue

	    #print parts
	    payload=''.join([p[0] for p in parts])
	    #print 'combined payload',payload

	    # Try to mirror the packet as much as possible... same seqId and channel
	    outStr = ','.join((fields[0],'1,1',fields[3],fields[4],payload, fields[6].split('*')[0]+'*'+checksumStr(payload),','.join(fields[7:])))
	    # FIX: Why do I have to do this last strip???
	    o.write(outStr.strip()+'\n')

	    continue

	buffers[bufferSlot].append(newPacket)



############################################################
if __name__=='__main__':
	from optparse import OptionParser
	parser = OptionParser(usage="%prog [options] file1.ais [file2.ais ...]",version="%prog "+__version__)

	parser.add_option('-o','--outfile',dest='outFile',default=None,#default='normalized.ais',
			  help='Name of the AIS file to write [default: stdout]')

	parser.add_option('-v','--verbose',dest='verbose',default=False,action='store_true',
		help='Make the test output verbose')

	(options,args) = parser.parse_args()

	out = sys.stdout
	if options.outFile != None: 
	    out = file(options.outFile,'w')
	
	# FIX: if no args, then pull from stdin
	if len(args)==0:
	    assembleAisNmeaMessages(sys.stdin,out)
	else:
	    for filename in args:
		assembleAisNmeaMessages(file(filename),out)

